import telegram

from telegram_django_bot.routing import telegram_resolve, telegram_reverse, RouterCallbackMessageCommandHandler
from telegram_django_bot.test import TD_TestCase
from telegram_django_bot.models import BotMenuElem, ActionLog
from django.urls.exceptions import NoReverseMatch
from django.conf import settings

# import os, sys
# sys.path.append(os.path.dirname(__file__))
# print(sys.path, os.path.dirname(__file__))

from test_app.models import User, Category
from test_app.views import CategoryViewSet, EntityViewSet
from test_app.handlers import me
import unittest

class TestTelegramResolve(TD_TestCase):
    def test_exist(self):
        res = telegram_resolve('cat/se')
        self.assertEqual(CategoryViewSet, res.func)

        res = telegram_resolve('/cat/', utrl_conf='test_app.utrls')
        self.assertEqual(CategoryViewSet, res.func)

        res = telegram_resolve('cat/up&1&2&3')
        self.assertEqual(CategoryViewSet, res.func)

        res = telegram_resolve('ent/cr&some_name&cat/')
        self.assertEqual(EntityViewSet, res.func)

        res = telegram_resolve('start')
        self.assertEqual(me, res.func)


    def test_not_exist(self):
        self.assertIsNone(telegram_resolve('abracadabra'))
        self.assertIsNone(telegram_resolve('cat'))
        self.assertIsNone(telegram_resolve('startstart/'))



class TestTelegramReverse(TD_TestCase):
    def test_exist(self):
        res = telegram_reverse('CategoryViewSet')
        self.assertEqual('cat/', res)

        res = telegram_reverse('EntityViewSet')
        self.assertEqual('ent/', res)

        res = telegram_reverse('self_info')
        self.assertEqual('me', res)

    def test_not_exist(self):
        catched = False
        try:
            telegram_reverse('abracadabra')
        except NoReverseMatch:
            catched = True

        self.assertTrue(catched)



@unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
class TestRouterCallbackMessageCommandHandler(TD_TestCase):
    def setUp(self) -> None:
        self.bme = BotMenuElem.objects.create(
            message='test',
            command='command'

        )

    # 'dispatcher', 'check_result' -- unnecessary args
    def test_command(self):
        rc_mch = RouterCallbackMessageCommandHandler()
        update = self.create_update({'text': '/start'})
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(telegram.Message, type(res))
        self.assertEqual(1, ActionLog.objects.filter(type='me').count())


    def test_bme_command(self):
        rc_mch = RouterCallbackMessageCommandHandler()
        update = self.create_update({'text': '/test'})
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(telegram.Message, type(res))

        rc_mch = RouterCallbackMessageCommandHandler(only_utrl=True)
        self.assertFalse(rc_mch.check_update(update))
        self.assertEqual(1, ActionLog.objects.filter(type='/test').count())


    def test_command_not_exist(self):
        rc_mch = RouterCallbackMessageCommandHandler()
        update = self.create_update({'text': '/abra'})
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(
            'Oops! It seems that an error has occurred, please write to support (contact in bio)!',
            res.text
        )
        self.assertEqual(1, ActionLog.objects.filter(type='/abra').count())


    def test_callback(self):
        rc_mch = RouterCallbackMessageCommandHandler()

        message = self.test_callback_context.bot.send_message(
            chat_id=settings.TELEGRAM_TEST_USER_IDS[0],
            text='some text'
        )
        callback_params = {'data': 'start'}

        update = self.create_update(message.to_dict(), callback_params)
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(1, ActionLog.objects.filter(type='me').count())
        self.assertEqual(telegram.Message, type(res))

    def test_callback_viewset(self):
        rc_mch = RouterCallbackMessageCommandHandler()

        message = self.test_callback_context.bot.send_message(
            chat_id=settings.TELEGRAM_TEST_USER_IDS[0],
            text='some text'
        )
        callback_params = {'data': 'cat/cr'}

        update = self.create_update(message.to_dict(), callback_params)
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(1, ActionLog.objects.filter(type='cat/cr').count())
        self.assertEqual(telegram.Message, type(res))

    def test_callback_user_utrl(self):
        rc_mch = RouterCallbackMessageCommandHandler()
        update = self.create_update({'text': 'category_name'})
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        user = User.objects.create(
            id=user_id,
            username=user_id,
            current_utrl='cat/cr&name'
        )

        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)

        self.assertEqual(1, Category.objects.count())
        self.assertEqual(1, ActionLog.objects.filter(type='cat/cr').count())
        self.assertEqual(telegram.Message, type(res))

    def test_callback_bme(self):
        rc_mch = RouterCallbackMessageCommandHandler()

        message = self.test_callback_context.bot.send_message(
            chat_id=settings.TELEGRAM_TEST_USER_IDS[0],
            text='some text'
        )
        callback_params = {'data': 'test'}
        update = self.create_update(message.to_dict(), callback_params)
        self.assertTrue(rc_mch.check_update(update))

        res = rc_mch.handle_update(update, 'dispatcher', 'check_result', self.test_callback_context)
        self.assertEqual(1, ActionLog.objects.filter(type='test').count())
        self.assertEqual(telegram.Message, type(res))

        rc_mch = RouterCallbackMessageCommandHandler(only_utrl=True)
        self.assertFalse(rc_mch.check_update(update))
        self.assertEqual(1, ActionLog.objects.filter(type='test').count())





import datetime

from telegram_django_bot.test import DJ_TestCase
from test_app.models import User
from django.conf import settings


class TestUser(DJ_TestCase):
    def setUp(self) -> None:
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        self.user1 = User.objects.create(id=user_id, username=user_id)
        self.user2 = User.objects.create(id=1, username='1')

    def test_clear_status(self):
        self.user1.current_utrl = 'aa/bb'
        self.user1.current_utrl_context_db = '{"a": "b"}'
        self.user1.current_utrl_form_db = '{"some": "info"}'
        self.user1.save()

        self.user1.clear_status(commit=False)
        self.assertEqual('', self.user1.current_utrl)
        self.assertEqual('{}', self.user1.current_utrl_context_db)
        self.assertEqual('{}', self.user1.current_utrl_form_db)

        self.user1.refresh_from_db()
        self.assertEqual('aa/bb', self.user1.current_utrl)
        self.assertEqual('{"a": "b"}', self.user1.current_utrl_context_db)

        self.user1.clear_status()
        self.user1.refresh_from_db()
        self.assertEqual('', self.user1.current_utrl)
        self.assertEqual('{}', self.user1.current_utrl_context_db)
        self.assertEqual('{}', self.user1.current_utrl_form_db)

    def test_save_and_load_form(self):
        form_name = 'TestForm'
        cleaned_data = {
            'int': 1000,
            'float': 3.14,
            'dt': datetime.date(2000, 1, 1),
            'tm': datetime.time(10, 30),
            'dttm': datetime.datetime(2023, 1, 1, 0, 45, 10),
            'user': self.user1,
            'several_models': User.objects.all().order_by('id'),
            'string': 'abrakadabra <b> 123:00 !?',
            'list': [1, 2, 3, 'a'],
            'dict': {'1': 2, '2': 3}
        }

        self.user1.save_form_in_db(form_name, cleaned_data)
        self.assertEqual(form_name, self.user1.current_utrl_form['form_name'])

        data_for_form = cleaned_data.copy()  # for form this change is okey
        data_for_form['several_models'] = ['1', str(self.user1.pk)]
        data_for_form['user'] = self.user1.pk
        self.assertEqual(data_for_form, self.user1.current_utrl_form['form_data'])

        self.user1.save_form_in_db('TestForm2', {})
        self.assertEqual({}, self.user1.current_utrl_form['form_data'])

    def test_current_utrl_context(self):
        data = [1, 2, {'a': {'dttm': datetime.datetime(2020, 1, 1, 1)}}]
        self.user1.save_context_in_db(data)
        self.user1.save()
        self.assertEqual(data, self.user1.current_utrl_context)

        data2 = {'a': 1}
        self.user1.save_context_in_db(data2)
        self.assertEqual(data2, self.user1.current_utrl_context)

    def test_language_selection(self):
        self.assertEqual('en', self.user1.language_code)

        self.user1.telegram_language_code = 'ru'
        self.assertEqual('ru', self.user1.language_code)

        self.user1.telegram_language_code = 'de'
        self.assertEqual('de', self.user1.language_code)

        self.user1.telegram_language_code = 'fr'  # not exist in settings
        self.assertEqual('en', self.user1.language_code)



from telegram_django_bot.test import TD_TestCase
from telegram_django_bot.telegram_lib_redefinition import InlineKeyboardButtonDJ, InlineKeyboardMarkupDJ
from telegram_django_bot.models import BotMenuElem, BotMenuElemAttrText, MESSAGE_FORMAT
from django.conf import settings

from test_app.models import User

import unittest
import telegram


@unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
class Test_TG_DJ_BOT(TD_TestCase):

    def setUp(self) -> None:
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        self.user1 = User.objects.create(id=user_id, username=user_id)

        self.user2 = User.objects.create(id=1, username=1)


    def test_send_format_message(self):
        bot = self.test_callback_context.bot
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]

        message, __ = bot.send_format_message(
            text='simple message',
            chat_id=user_id
        )

        callback_params = {'data': 'just_for_update'}
        update = self.create_update({'message_id': message.message_id}, callback_params)

        message, __ = bot.send_format_message(
            text='simple <b> message',  # without closed </b> -- checked that parse_mode is changed
            update=update,
            reply_markup=InlineKeyboardMarkupDJ([[InlineKeyboardButtonDJ('test', callback_data='test')]]),
            disable_web_page_preview=True,
            parse_mode='Markdown'
        )

        photo1 = open('media/pic1.png', 'rb')
        update = self.create_update(message.to_dict(), callback_params)
        message, __ = bot.send_format_message(
            message_format=MESSAGE_FORMAT.PHOTO,
            media_files_list=[photo1],
            update=update,
        )

        # photo1 = InputMediaPhoto(open('media/pic1.png', 'rb'))
        # photo2 = InputMediaPhoto(open('media/pic2.jpg', 'rb'))
        # update = self.create_update(message.to_dict(), callback_params)
        # message, __ = bot.send_format_message(
        #     message_format=MESSAGE_FORMAT.GROUP_MEDIA,
        #     media_files_list=[photo1, photo2],
        #     update=update,
        # )

        doc = open('media/game.gif', 'rb')
        message, __ = bot.send_format_message(
            text='test',
            chat_id=user_id,
            message_format=MESSAGE_FORMAT.DOCUMENT,
            media_files_list=[doc],
            disable_web_page_preview=True,
        )

        # todo: add check for all formats

    def test_send_botmenuelem(self):
        bme = BotMenuElem.objects.create(
            message_format=MESSAGE_FORMAT.PHOTO,
            message='bme send test',
            buttons_db='[[{"text":"button", "url": "google.com"}]]',
            media='media/pic1.png'
        )
        BotMenuElemAttrText.objects.filter(
            bot_menu_elem=bme,
            default_text='bme send test',
            language_code='ru'
        ).update(translated_text='bme отправленное сообщение')

        bot = self.test_callback_context.bot

        message = bot.send_botmenuelem(None, self.user1, None)
        self.assertEqual(
            'Oops! It seems that an error has occurred, please write to support (contact in bio)!',
            message.text,
        )
        self.assertIsNone(bme.telegram_file_code)

        message = bot.send_botmenuelem(None, self.user1, bme)
        self.assertEqual(
            'bme send test',
            message.caption,
        )
        self.assertIsNotNone(bme.telegram_file_code)

        self.user1.telegram_language_code = 'ru'
        self.user1.save()
        update = self.create_update(message.to_dict(), {'data': 'just_for_update'})
        message = bot.send_botmenuelem(update, self.user1, bme)
        self.assertEqual(
            'bme отправленное сообщение',
            message.caption,
        )

        bme.refresh_from_db()



    def test_edit_or_send(self):
        bot = self.test_callback_context.bot
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]

        message = bot.edit_or_send(
            None,
            'test mess',
            chat_id=user_id
        )

        callback_params = {'data': 'just_for_update'}

        update = self.create_update(message.to_dict(), callback_params)
        message = bot.edit_or_send(
            update,
            'test mess 2',
            buttons=[],
        )

        update = self.create_update(message.to_dict(), callback_params)
        message = bot.edit_or_send(
            update,
            'test mess 3',
            buttons=[
                [InlineKeyboardButtonDJ('button 1', url='google.com')],
                [
                    InlineKeyboardButtonDJ('button 2', callback_data='aaa'),
                    InlineKeyboardButtonDJ('button 3 😄', callback_data='bbb'),
                ]
            ],
        )

        update = self.create_update(message.to_dict())
        message = bot.edit_or_send(
            update,
            'test mess 4',
            reply_to_message_id=message.message_id,
        )


        update = self.create_update(message.to_dict(), callback_params)
        message = bot.edit_or_send(
            update,
            'test mess <b> 5',
            disable_notification=True,
            disable_web_page_preview=True,
            parse_mode='Markdown',
            timeout=20,
            only_send=True,
        )

    def test_task_send_message_handler(self):
        bot = self.test_callback_context.bot

        func_args = []
        func_kwargs = {
            'text': 'test',
            'chat_id': self.user1.id,
        }

        is_sent, res_mess = bot.task_send_message_handler(
            self.user1,
            bot.send_format_message,
            func_args,
            func_kwargs,
        )
        self.assertTrue(is_sent)

        func_kwargs['chat_id'] = self.user2.id
        is_sent, res_mess = bot.task_send_message_handler(
            self.user2,
            bot.send_format_message,
            func_args,
            func_kwargs,
        )
        self.assertFalse(is_sent)
        self.assertFalse(self.user2.is_active)


        # add translation check


# Generated by Django 3.1 on 2022-12-25 10:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('telegram_django_bot', '0007_auto_20221202_0305'),
    ]

    operations = [
        migrations.AlterField(
            model_name='botmenuelemattrtext',
            name='language_code',
            field=models.CharField(max_length=16),
        ),
        migrations.AlterField(
            model_name='botmenuelem',
            name='buttons_db',
            field=models.TextField(default='[]',
                                   help_text='InlineKeyboardMarkup buttons list, ({"text": , "url" or "callback_data"})'),
        ),
        migrations.AlterField(
            model_name='botmenuelem',
            name='callbacks_db',
            field=models.TextField(default='[]',
                                   help_text='List of regular expressions (so far only an explicit list) for callbacks that call this menu block'),
        ),
        migrations.AlterField(
            model_name='botmenuelem',
            name='buttons_db',
            field=models.TextField(default='[]',
                                   help_text='InlineKeyboardMarkup buttons structure (double list of dict), where each button(dict) has next format: {"text": "text", "url": "google.com"} or {"text":"text", "callback_data": "data"})'),
        ),
        migrations.AlterField(
            model_name='botmenuelem',
            name='command',
            field=models.TextField(blank=True, help_text='Bot command that can call this menu block', null=True),
        ),
    ]



import json
from telegram import (
    # TelegramObject,
    Bot,
    InlineKeyboardMarkup,
    InlineKeyboardButton,
    KeyboardButton,
)


try:
    # version 20.x +
    from telegram import ReplyKeyboardMarkup as ReplyMarkup
except ImportError:
    # old version
    from telegram import ReplyMarkup


def dj_translate_obj_2_string(obj):
    if issubclass(type(obj), object) and obj.__class__.__name__ == '__proxy__':
        return str(obj)

    elif hasattr(obj, '__call__'):
        return obj.__call__()

    return None # if it is not django translate object


class TelegramDjangoEncoder(json.JSONEncoder):
    def default(self, o):
        res = dj_translate_obj_2_string(o)
        if res is None:
            return super(TelegramDjangoEncoder, self).default(o)
        return res


class TelegramDjangoObject2Json:
    __slots__ = ()

    def to_json(self) -> str:
        # for old versions
        return json.dumps(self.to_dict(), cls=TelegramDjangoEncoder)


    def _get_attrs(self, *args, **kwargs):
        # for new 20.x+ versions

        data = super()._get_attrs(*args, **kwargs)
        for key, o in data.items():
            res = dj_translate_obj_2_string(o)
            if res is not None:
                data[key] = res
        return data



class InlineKeyboardMarkupDJ(TelegramDjangoObject2Json, InlineKeyboardMarkup):
    pass


class InlineKeyboardButtonDJ(TelegramDjangoObject2Json, InlineKeyboardButton):
    pass


class ReplyMarkupDJ(TelegramDjangoObject2Json, ReplyMarkup):
    pass


class KeyboardButtonDJ(TelegramDjangoObject2Json, KeyboardButton):
    pass


class BotDJ(TelegramDjangoObject2Json, Bot):
    def _check_django_localization(self, data):
        '''
        check 'text', 'caption' attributes for translation
        :param data:
        :return:
        '''
        for field in ['text', 'caption']:
            if field in data and hasattr(data[field], '__call__'):
                data[field] = data[field].__call__()
            elif data.get(field):
                data[field] = str(data[field])
        return data


    def _message(
            self,
            endpoint: str,
            data,
            *args,
            **kwargs
    ):
        """
        this method used in old versions.

        """

        data = self._check_django_localization(data)

        return super()._message(
            endpoint,
            data,
            *args,
            **kwargs
        )

    async def _do_post(
        self,
        endpoint: str,
        data,
        *args,
        **kwargs,
    ):
        """
        this method used in new 20.x+ versions.

        """

        data = self._check_django_localization(data)
        return await super()._do_post(endpoint, data, *args, **kwargs)




from telegram_django_bot.test import TD_TestCase
from telegram_django_bot.models import ActionLog, TeleDeepLink, BotMenuElem
from telegram_django_bot.utils import handler_decor
from django.conf import settings
from test_app.models import User

import unittest
import telegram


@handler_decor(log_type='C')
def normal_func(bot, update, user):
    return


@handler_decor(log_type='F')
def func_with_error(bot, update, user):
    raise ArithmeticError('error :(')


class TestHandlerDecor(TD_TestCase):
    def setUp(self) -> None:
        self.user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        self.update = self.create_update({'text': '/start deeplink'})

    def test_user_creating(self):
        normal_func(self.update, self.test_callback_context)

        self.assertEqual(self.user_id, User.objects.filter().get().id)
        self.assertEqual(1, TeleDeepLink.objects.filter(link='deeplink').count())
        self.assertEqual(3, ActionLog.objects.count())
        self.assertEqual(1, ActionLog.objects.filter(user_id=self.user_id, type='/start deeplink').count())

    def test_for_exist_user(self):
        User.objects.create(id=self.user_id, username=self.user_id)

        normal_func(self.update, self.test_callback_context)
        self.assertEqual(0, TeleDeepLink.objects.all().count())
        self.assertEqual(2, ActionLog.objects.count())

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_error(self):
        error = None
        try:
            func_with_error(self.update, self.test_callback_context)
        except ArithmeticError as err:
            error = err

        self.assertIsNotNone(error)
        self.assertEqual(3, ActionLog.objects.count())
        self.assertEqual(1, ActionLog.objects.filter(user_id=self.user_id, type='func_with_error').count())


    # def test_language(self):
    #     # todo:
    #     pass


import logging

from django.test import TestCase
from django.conf import settings
from django.utils import timezone
from telegram import Update, Message, Chat, User as TelegramAPIUser, CallbackQuery
from telegram_django_bot.routing import RouterCallbackMessageCommandHandler
# from telegram.utils.types import JSONDict, ODVInput
# from telegram.utils.helpers import DEFAULT_NONE
# from typing import Union

from .tg_dj_bot import TG_DJ_Bot
# import asyncio


test_bot = TG_DJ_Bot(token=settings.TELEGRAM_TOKEN)
if hasattr(settings, 'TELEGRAM_TEST_USER_IDS'):
    TELEGRAM_TEST_USER_IDS = settings.TELEGRAM_TEST_USER_IDS
    for user_id in TELEGRAM_TEST_USER_IDS:
        try:
            res = test_bot.send_message(user_id, 'ping')

            # if asyncio.iscoroutine(res):
            #     asyncio.run(res)

        except Exception as error:
            logging.error(
                f'the error {error} occurred while sending test message to TELEGRAM_TEST_USER_ID={user_id} '
                f'by bot {test_bot.id}. Please, check details'
            )
else:
    TELEGRAM_TEST_USER_IDS = []
    logging.warning('TELEGRAM_TEST_USER_IDS is not set in settings. Set it or specify user in each test functions')


class TestCallbackContext:
    bot = test_bot


class DJ_TestCase(TestCase):
    pass


class TD_TestCase(DJ_TestCase):
    # test_bot = test_bot
    # test_callback_context = TestCallbackContext()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.test_callback_context = TestCallbackContext()
        self.rc_mch = RouterCallbackMessageCommandHandler()
        self.handle_update = lambda update: self.rc_mch.handle_update(
            update, 'some_str', 'some_str', self.test_callback_context
        )

    def create_update(self, message_kwargs: dict = None, callback_kwargs: dict = None, user_id=None,):
        if user_id is None and len(TELEGRAM_TEST_USER_IDS):
            user_id = TELEGRAM_TEST_USER_IDS[0]

        if user_id is None:
            raise ValueError('the user_id (or TELEGRAM_TEST_USER_IDS) should be set for creating Update model')

        chat = Chat(id=user_id, type='private')
        user = TelegramAPIUser(id=user_id, first_name='test', is_bot=False)

        message_kwargs = dict(message_kwargs) if message_kwargs else {}
        message_kwargs['message_id'] = message_kwargs.get('message_id', 1)
        message_kwargs['date'] = message_kwargs.get('message_id', timezone.now())
        message_kwargs['chat'] = message_kwargs.get('chat', chat)
        message_kwargs['from_user'] = message_kwargs.get('from_user', user)
        message = Message(**message_kwargs)

        if callback_kwargs:
            callback_kwargs = dict(callback_kwargs)
            callback_kwargs['id'] = callback_kwargs.get('id', 1)
            callback_kwargs['from_user'] = callback_kwargs.get('from_user', user)
            callback_kwargs['message'] = callback_kwargs.get('message', message)
            callback_kwargs['chat_instance'] = callback_kwargs.get('chat_instance', '1')
            callback = CallbackQuery(**callback_kwargs)
        else:
            callback = None

        update_kwargs = {
            'update_id': 1
        }
        if callback:
            update_kwargs['callback_query'] = callback
        else:
            update_kwargs['message'] = message

        return Update(**update_kwargs)










import time
import sys
import logging
from telegram import (
    Update,
    InputMediaPhoto,
    InputMediaVideo,
    InputMediaAudio,
    InputMediaAnimation,
    InputMediaDocument,
    Message,
    error,
)
from django.conf import settings  # LANGUAGES, USE_I18N
from django.utils import translation

from .models import BotMenuElem, MESSAGE_FORMAT
from .utils import add_log_action, ERROR_MESSAGE
from .telegram_lib_redefinition import (
    InlineKeyboardMarkupDJ,
    BotDJ,
    # TelegramDjangoObject2Json,
)


class TG_DJ_Bot(BotDJ):
    """
    no_error_send
    bot_menu_elem
    """

    def edit_or_send(bot, update, mess, buttons=None, only_send=False, decorate_buttons=True, **kwargs):
        if decorate_buttons and buttons:
            marked_buttons = InlineKeyboardMarkupDJ(buttons)
        else:
            marked_buttons = buttons

        if not marked_buttons:
            marked_buttons = None

        kwargs['reply_markup'] = marked_buttons

        res_mess, _ = bot.send_format_message(
            MESSAGE_FORMAT.TEXT,
            mess,
            update=update,
            only_send=only_send,
            **kwargs,
        )
        return res_mess

    def send_botmenuelem(bot, update, user, menu_elem):
        if menu_elem is None:
            menu_elem = BotMenuElem.objects.filter(empty_block=True, is_visable=True).first()

        media_files_list = None
        extra_kwargs = {}

        if menu_elem is None:
            if settings.USE_I18N:
                translation.activate(user.language_code)

            message_format = MESSAGE_FORMAT.TEXT
            mess = str(ERROR_MESSAGE)

            utrl = ''
            if update:
                if update.message:
                    utrl = update.message.text
                elif update.callback_query:
                    utrl = update.callback_query.data

            logging.warning(
                f'Try to find BME {utrl}, but there is no such models, so the error message shown for the client'
            )
        else:
            language_code = settings.LANGUAGE_CODE
            if settings.USE_I18N and user.language_code != settings.LANGUAGE_CODE:
                language_code = user.language_code

            message_format = menu_elem.message_format
            mess = menu_elem.get_message(language_code)
            buttons = menu_elem.get_buttons(language_code)

            if len(buttons):
                extra_kwargs['reply_markup'] = InlineKeyboardMarkupDJ(buttons)

            if menu_elem.message_format != MESSAGE_FORMAT.TEXT:
                media_file = menu_elem.telegram_file_code or open(menu_elem.media.path, 'rb')

                media_files_list = [media_file]

        response, media_codes = bot.send_format_message(
            message_format,
            mess,
            media_files_list,
            update,
            user.id,

            **extra_kwargs,
        )

        if menu_elem and menu_elem.telegram_file_code is None and menu_elem.media and len(media_codes) > 0:
            menu_elem.telegram_file_code = media_codes[0]
            menu_elem.save()
        return response

    def send_format_message(
            bot,
            message_format: str = MESSAGE_FORMAT.TEXT,
            text: str = None,
            media_files_list: list = None,
            update: Update = None,
            chat_id: int = None,
            only_send=False,

            **telegram_message_kwargs

    ):
        """

        :param message_format:
        :param text:
        :param media_files_list: for send_media_group must be prepared with InputMediaPhoto, InputMediaVideo, InputMediaDocument, InputMediaAudio
        :param update:
        :param chat_id:
        :param only_send: if true then send new message. if callback_query is exist -- then delete previous message


        telegram_message_kwargs - parse_mode, timeout, reply_markup, disable_web_page_preview, disable_notification and maybe other
        :return:
        """

        input_media_dict = {
            MESSAGE_FORMAT.PHOTO: InputMediaPhoto,
            MESSAGE_FORMAT.VIDEO: InputMediaVideo,
            MESSAGE_FORMAT.AUDIO: InputMediaAudio,
            MESSAGE_FORMAT.GIF: InputMediaAnimation,
            MESSAGE_FORMAT.DOCUMENT: InputMediaDocument,
        }

        # checks data
        if message_format == MESSAGE_FORMAT.TEXT:
            if text is None:
                raise ValueError(f'text message could not be without text: got message_format={message_format}, text={text}')
        elif media_files_list is None:
            raise ValueError(f'not text message without media: got message_format={message_format}, media_files_list={media_files_list}')
        elif type(media_files_list) != list:
            raise ValueError(f'media_files_list should be list, got {type(media_files_list)}: {media_files_list}')
        elif len(media_files_list) == 0:
            raise ValueError(f'not text message without media: got message_format={message_format}, media_files_list={media_files_list}')
        elif message_format != MESSAGE_FORMAT.GROUP_MEDIA and len(media_files_list) > 1:
            raise ValueError(f'not group media message with more then 1 file: got message_format={message_format}, media_files_list={media_files_list}')
        elif message_format == MESSAGE_FORMAT.GROUP_MEDIA and (len(media_files_list) < 2 or len(media_files_list) > 10):
            raise ValueError(f'group media message should be with [2, 10] media, got {len(media_files_list)}')

        if update is None and chat_id is None:
            raise ValueError(f'update and chat_id could be together None -- to which chat send message then?')

        chat_id = chat_id or update.effective_user.id
        # text = str(text)  # for internalization (django __proxy__ error)

        telegram_message_kwargs = telegram_message_kwargs.copy()
        if not 'parse_mode' in telegram_message_kwargs:
            telegram_message_kwargs['parse_mode'] = 'HTML'

        delete_message_id = None
        is_editing_message = False
        if update and update.callback_query:
            prev_mess = update.callback_query.message
            if only_send:
                delete_message_id = prev_mess.message_id
            elif prev_mess.text         and message_format != MESSAGE_FORMAT.TEXT or \
                 prev_mess.text is None and message_format == MESSAGE_FORMAT.TEXT:
                # prev_mess.document and message_format != MESSAGE_FORMAT.DOCUMENT or \
                #  prev_mess.audio and message_format != MESSAGE_FORMAT.AUDIO or \
                #  prev_mess.photo and message_format != MESSAGE_FORMAT.PHOTO or \
                #  prev_mess.video and message_format != MESSAGE_FORMAT.VIDEO or \
                #  prev_mess.video and message_format != MESSAGE_FORMAT.VIDEO or \
                #  prev_mess.animation and message_format != MESSAGE_FORMAT.GIF or \
                #  prev_mess.text and message_format != MESSAGE_FORMAT.TEXT or \
                #  prev_mess.media_group_id:

                delete_message_id = prev_mess.message_id
            else:
                is_editing_message = True

        if delete_message_id:
            bot.delete_message(chat_id, delete_message_id)

        if is_editing_message:
            edit_message_id = update.callback_query.message.message_id
            telegram_message_kwargs.pop('disable_notification', None)
            if message_format == MESSAGE_FORMAT.TEXT:
                response = bot.edit_message_text(
                    text,
                    chat_id,
                    message_id=edit_message_id,
                    **telegram_message_kwargs
                )
            else:
                telegram_message_kwargs.pop('disable_web_page_preview', None)

                input_media = input_media_dict[message_format]

                media_file = input_media(
                    media_files_list[0],
                    caption=text,
                    parse_mode=telegram_message_kwargs.pop('parse_mode', None)  # fixme: why pop?
                )

                response = bot.edit_message_media(
                    chat_id,
                    message_id=edit_message_id,
                    media=media_file,
                    **telegram_message_kwargs
                )
        else:
            if message_format == MESSAGE_FORMAT.TEXT:
                response = bot.send_message(
                    chat_id,
                    text,
                    **telegram_message_kwargs
                )
            else:
                telegram_message_kwargs.pop('disable_web_page_preview', None)

                if message_format == MESSAGE_FORMAT.GROUP_MEDIA:
                    telegram_message_kwargs.pop('reply_markup', None)
                    response = bot.send_media_group(
                        chat_id,
                        media=media_files_list,

                    )
                else:
                    if message_format == MESSAGE_FORMAT.PHOTO:
                        telegram_func = bot.send_photo
                    elif message_format == MESSAGE_FORMAT.AUDIO:
                        telegram_func = bot.send_audio
                    elif message_format == MESSAGE_FORMAT.VIDEO:
                        telegram_func = bot.send_video
                    elif message_format == MESSAGE_FORMAT.GIF:
                        telegram_func = bot.send_animation
                    elif message_format == MESSAGE_FORMAT.VIDEO_NOTE:
                        telegram_func = bot.send_video_note
                    elif message_format == MESSAGE_FORMAT.VOICE:
                        telegram_func = bot.send_voice
                    elif message_format == MESSAGE_FORMAT.STICKER:
                        telegram_func = bot.send_sticker
                    elif message_format == MESSAGE_FORMAT.LOCATION:
                        raise NotImplementedError()
                    # elif message_format == MESSAGE_FORMAT.DOCUMENT:
                    else:
                        telegram_func = bot.send_document

                    response = telegram_func(
                        chat_id,
                        media_files_list[0],
                        caption=text,
                        **telegram_message_kwargs
                    )

        media_files_codes = []
        if type(response) == Message:
            media_file = response.document or response.audio or response.video or response.animation
            if media_file is None and response.photo:
                media_file = response.photo[-1] # last -- biggest
            if media_file:
                media_files_codes.append(media_file.file_id)

            # todo: add support for group media

        return response, media_files_codes

    def task_send_message_handler(bot, user, func, func_args, func_kwargs):
        is_sent = False
        res_mess = None
        try:
            if settings.USE_I18N:
                translation.activate(user.language_code)

            res_mess = func(*func_args, **func_kwargs)
            is_sent = True
            time.sleep(0.035)

        except error.Unauthorized as e:
            logging.info(f'user blocked {e}, {user.id}\n')
            user.is_active = False
            user.save()
            add_log_action(user.id, 'TYPE_BLOCKED')

        except error.BadRequest as ee:
            if ee.message == 'Chat not found':
                user.is_active = False
                user.save()
                add_log_action(user.id, 'TYPE_BLOCKED')

            else:
                logging.error(
                    f'{ee.with_traceback(sys.exc_info()[2])} \n user={user.id}, {func}, {func_args}, {func_kwargs}'
                )

        except Exception as ee:
            logging.error(
                f'{ee.with_traceback(sys.exc_info()[2])} \n user={user.id}, {func}, {func_args}, {func_kwargs}'
            )
            time_in_seconds = 0.4
            try:
                if 'Flood control exceeded. Retry in ' in str(ee):
                    time_in_seconds = str(ee)[len('Flood control exceeded. Retry in '):].split()[0]
                    time_in_seconds = float(time_in_seconds)

                time.sleep(time_in_seconds)
            except Exception as e:
                time.sleep(0.5)

        return is_sent, res_mess





import logging

from django.db import models
from django.contrib.auth.models import AbstractUser
from django.core import validators
import random
from django.utils import timezone
import json
from django.conf import settings
from django.db.models import QuerySet
from django.core.serializers.json import DjangoJSONEncoder
import datetime
from django.utils.translation import gettext_lazy as _
from telegram import InlineKeyboardButton  # no lazy text so standart possible to use


class TelegramDjangoJsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super(TelegramDjangoJsonDecoder, self).__init__(*args, object_hook=self.object_hook_decoder, **kwargs)

    def object_hook_decoder(self, sub_dict, *args, **kwargs):
        # so it works only for dictionary data (if datetime is in list it will not be worked)
        for key in sub_dict.keys():
            value = sub_dict[key]
            dt_object = None
            if type(value) == str:
                try:
                    dt_object = datetime.time.fromisoformat(value)
                except ValueError:
                    try:
                        dt_object = datetime.date.fromisoformat(value)
                    except ValueError:
                        try:
                            dt_object = datetime.datetime.fromisoformat(value)
                        except ValueError:
                            pass

                if dt_object:
                    sub_dict[key] = dt_object

        return sub_dict


class MESSAGE_FORMAT:
    TEXT = 'T'
    PHOTO = 'P'
    DOCUMENT = 'D'
    AUDIO = 'A'
    VIDEO = 'V'
    GIF = 'G'
    VOICE = 'TV'
    VIDEO_NOTE = 'VN'
    STICKER = 'S'
    LOCATION = 'L'
    GROUP_MEDIA = 'GM'

    MESSAGE_FORMATS = (
        (TEXT, _('Text')),
        (PHOTO, _('Image')),
        (DOCUMENT, _('Document')),
        (AUDIO, _('Audio')),
        (VIDEO, _('Video')),
        (GIF, _('GIF/animation')),
        (VOICE, _('Voice')),
        (VIDEO_NOTE, _('Video note')),
        (STICKER, _('Sticker')),
        (LOCATION, _('Location')),
        (GROUP_MEDIA, _('Media Group')),
    )

    ALL_FORMATS = (elem[0] for elem in MESSAGE_FORMATS)


class ModelwithTimeManager(models.Manager):
    def bot_filter_active(self, *args, **kwargs):
        return self.filter(dttm_deleted__isnull=True, *args, **kwargs)


class AbstractActiveModel(models.Model):
    dttm_added = models.DateTimeField(default=timezone.now)
    dttm_deleted = models.DateTimeField(null=True, blank=True)

    objects = ModelwithTimeManager()

    class Meta:
        abstract = True


class TelegramAbstractActiveModel(AbstractActiveModel):
    message_id = models.BigIntegerField(null=True, blank=True)

    class Meta:
        abstract = True


def _seed_code():
    return random.randint(0, 100)


class TelegramUser(AbstractUser):
    id = models.BigIntegerField(primary_key=True)  # telegram id is id for models
    seed_code = models.IntegerField(default=_seed_code)
    telegram_username = models.CharField(max_length=64, null=True, blank=True)
    telegram_language_code = models.CharField(max_length=16, default='en')  # could be with dialects

    timezone = models.DurationField(default=timezone.timedelta(hours=3))

    current_utrl = models.CharField(max_length=64, default='', blank=True) # todo: add verify comparisson current_utrl and current_utrl_context_db/current_utrl_form_db
    current_utrl_code_dttm = models.DateTimeField(null=True, blank=True)
    current_utrl_context_db = models.CharField(max_length=4096, default='{}', blank=True)

    # form structure {'form_name': '', 'form_data': {}}
    current_utrl_form_db = models.CharField(max_length=4096, default='{}', blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return f"U({self.id}, {self.telegram_username or '-'}, {self.first_name or '-'})"

    @property
    def current_utrl_form(self):
        if not hasattr(self, '_current_utrl_form'):
            self._current_utrl_form = json.loads(self.current_utrl_form_db, cls=TelegramDjangoJsonDecoder)
        return self._current_utrl_form

    @property
    def current_utrl_context(self):
        if not hasattr(self, '_current_utrl_context'):
            self._current_utrl_context = json.loads(self.current_utrl_context_db, cls=TelegramDjangoJsonDecoder)
        return self._current_utrl_context

    def save_form_in_db(self, form_name, form_data, do_save=True):
        db_form_data = {}

        for key, value in form_data.items():
            if issubclass(value.__class__, models.Model):
                db_value = value.pk
            elif (type(value) in [list, QuerySet]) and all(map(lambda x: issubclass(x.__class__, models.Model), value)):
                db_value = list([str(x.pk) for x in value])
            else:
                db_value = value

            db_form_data[key] = db_value

        self.current_utrl_form_db = json.dumps({
            'form_name': form_name,
            'form_data': db_form_data,
        }, cls=DjangoJSONEncoder)
        if do_save:
            self.save()

        if hasattr(self, '_current_utrl_form'):
            delattr(self, '_current_utrl_form')

    def save_context_in_db(self, context, do_save=True):
        self.current_utrl_context_db = json.dumps(context, cls=DjangoJSONEncoder)
        if do_save:
            self.save()

        if hasattr(self, '_current_utrl_context'):
            delattr(self, '_current_utrl_context')

    def clear_status(self, commit=True):
        self.current_utrl = ''
        self.current_utrl_code_dttm = None
        self.current_utrl_context_db = '{}'
        self.current_utrl_form_db = '{}'
        if commit:
            self.save()

        for attr in ['_current_utrl_context', '_current_utrl_form']:
            if hasattr(self, attr):
                delattr(self, attr)


    @property
    def language_code(self):
        if self.telegram_language_code in map(lambda x: x[0], settings.LANGUAGES):
            return self.telegram_language_code
        return settings.LANGUAGE_CODE

    def save(self, *args, **kwargs):
        if self.id is None and self.is_staff:
            id_num = 1
            while self.__class__.objects.filter(id=id_num).count():
                id_num += 1

            self.id = id_num

            logging.warning(f"Try to save user without ID. For staff the smallest unused ID will be used: {id_num}")

        return super(TelegramUser, self).save(*args, **kwargs)
            

class TeleDeepLink(models.Model):
    title = models.CharField(max_length=64, default='', blank=True)
    price = models.FloatField(null=True, blank=True)
    link = models.CharField(max_length=64, validators=[validators.RegexValidator(
        '^[a-zA-Z0-9_-]+$',
        _('Telegram is only accepted letters, numbers and signs - _'),
    )])
    users = models.ManyToManyField(settings.AUTH_USER_MODEL)

    def __str__(self):
        return f'TDL({self.id}, {self.link})'

class ActionLog(models.Model):
    """
    User actions logs
    """

    dttm = models.DateTimeField(auto_now_add=True, db_index=True)
    type = models.CharField(max_length=64)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, blank=True)

    def __str__(self):
        return 'AL({}, {}, {})'.format(self.user_id, self.dttm, self.type)


class BotMenuElem(models.Model):
    """

    """

    command = models.TextField(  # for multichoice start
        null=True, blank=True,  # todo: add manual check
        help_text=_('Bot command that can call this menu block. Add 1 command per row')
    )

    empty_block = models.BooleanField(
        default=False,
        help_text=_('This block will be shown if there is no catching callback')
    )
    is_visable = models.BooleanField(
        default=True,
        help_text=_('Whether to display this menu block to users (can be hidden and not deleted for convenience)')
    )

    callbacks_db = models.TextField(
        default='[]',
        help_text=_(
            'List of regular expressions (so far only an explicit list) for callbacks that call this menu block. '
            'For example, list ["data", "callback2"] will catch the clicking InlineKeyboardButtons with callback_data "data" or "callback2"'
        )
    )

    forward_message_id = models.IntegerField(null=True, blank=True)
    forward_chat_id = models.IntegerField(null=True, blank=True)

    message_format = models.CharField(max_length=2, choices=MESSAGE_FORMAT.MESSAGE_FORMATS, default=MESSAGE_FORMAT.TEXT)
    message = models.TextField(help_text=_('Text message'))
    buttons_db = models.TextField(
        default='[]',
        help_text=_(
            'InlineKeyboardMarkup buttons structure (double list of dict), where each button(dict) has next format: '
            '{"text": "text", "url": "google.com"} or {"text":"text", "callback_data": "data"})'
        )
    )
    media = models.FileField(help_text=_('File attachment to the message'), null=True, blank=True)
    telegram_file_code = models.CharField(
        max_length=512, null=True, blank=True,
        help_text=_('File code in telegram (must be deleted when replacing file)')
    )

    def __str__(self):
        return f"BME({self.id}, { self.command[:32] if self.command else self.message[:32]})"

    def save(self, *args, **kwargs):
        # bot = telegram.Bot(TELEGRAM_TOKEN)
        
        super(BotMenuElem, self).save(*args, **kwargs)

        # check and create new models for translation
        if settings.USE_I18N and len(settings.LANGUAGES):
            language_codes = set(map(lambda x: x[0], settings.LANGUAGES))
            if settings.LANGUAGE_CODE in language_codes:
                language_codes.remove(settings.LANGUAGE_CODE)

            get_existed_language_codes = lambda text: set(BotMenuElemAttrText.objects.filter(
                language_code__in=language_codes,
                bot_menu_elem_id=self.id,
                default_text=text,
            ).values_list('language_code', flat=True))

            BotMenuElemAttrText.objects.bulk_create([
                BotMenuElemAttrText(language_code=language_code, default_text=self.message, bot_menu_elem_id=self.id)
                for language_code in language_codes - get_existed_language_codes(self.message)
            ])

            for row_elem in self.buttons:
                for elem in row_elem:
                    if text := elem.get('text'):
                        BotMenuElemAttrText.objects.bulk_create([
                            BotMenuElemAttrText(language_code=language_code, default_text=text, bot_menu_elem_id=self.id)
                            for language_code in language_codes - get_existed_language_codes(text)
                        ])

    @property
    def buttons(self):
        if not hasattr(self, '_buttons'):
            self._buttons = json.loads(self.buttons_db)

        return self._buttons

    @property
    def callbacks(self):
        if not hasattr(self, '_callbacks'):
            self._callbacks = json.loads(self.callbacks_db)

        return self._callbacks

    def get_message(self, language='en'):
        get_translate_model = None
        if language != settings.LANGUAGE_CODE:
            get_translate_model = BotMenuElemAttrText.objects.filter(
                language_code=language,
                bot_menu_elem_id=self.id,
                default_text=self.message,
                translated_text__isnull=False,
            ).first()
        text = get_translate_model.translated_text if get_translate_model else self.message
        return text

    def get_buttons(self, language='en'):
        # todo: rewrite with 1 request to db (default_text__in=...)

        get_translate_model = lambda text: BotMenuElemAttrText.objects.filter(
            language_code=language,
            bot_menu_elem_id=self.id,
            default_text=text,
            translated_text__isnull=False,
        ).first()

        need_translation = language != settings.LANGUAGE_CODE and settings.USE_I18N
        buttons = []

        for row_elem in self.buttons:
            row_buttons = []
            for item_in_row in row_elem:
                elem = dict(item_in_row)
                if elem.get('text') and need_translation and (translate_model := get_translate_model(elem['text'])):
                    elem['text'] = translate_model.translated_text
                row_buttons.append(InlineKeyboardButton(**elem))

            buttons.append(row_buttons)
        return buttons


class BotMenuElemAttrText(models.Model):
    class Meta:
        unique_together = [['bot_menu_elem', 'language_code', 'default_text']]
        index_together = [['bot_menu_elem', 'language_code', 'default_text']]

    dttm_added = models.DateTimeField(default=timezone.now)
    bot_menu_elem = models.ForeignKey(BotMenuElem, null=False, on_delete=models.CASCADE)

    language_code = models.CharField(max_length=16)
    default_text = models.TextField(
        help_text=_('The text which should be translate')
    )
    translated_text = models.TextField(null=True, help_text=_('Default_text Translation'))


class Trigger(AbstractActiveModel):

    name = models.CharField(max_length=512, unique=True)
    condition_db = models.TextField(help_text='''
    {
        seeds: [1, 2, 3, 4, 5],
        'amount': [{
            'gte': 5,
            'type__contains': 'dd',  // type__in, type
            'duration': '7d'
        }]
    }
    ''')

    min_duration = models.DurationField(
        help_text=_('the minimum period in which there can be 1 notification for a user of this type')
    )
    priority = models.IntegerField(default=1, help_text=_('the more topics will be executed first'))

    botmenuelem = models.ForeignKey(
        BotMenuElem, on_delete=models.PROTECT,
        help_text=_('which trigger message to show')
    )

    @property
    def condition(self):
        if not hasattr(self, '_condition'):
            self._condition = json.loads(self.condition_db)

        return self._condition

    @staticmethod
    def get_timedelta(delta_string:str):
        days = 0
        hours = 0
        for part in delta_string.split():
            if 'd' in part:
                days = float(part.replace('d', ''))
            elif 'h' in part:
                hours = float(part.replace('h', ''))
            else:
                raise ValueError(f'unknown format {part}')

        return timezone.timedelta(days=days, hours=hours)

    def __str__(self):
        return f"T({self.id}, {self.name})"


class UserTrigger(TelegramAbstractActiveModel):
    trigger = models.ForeignKey(Trigger, on_delete=models.PROTECT)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    is_sent = models.BooleanField(default=False)




import telegram
from telegram import InlineKeyboardButton, InlineKeyboardMarkup

from telegram_django_bot.test import TD_TestCase
from telegram_django_bot.models import BotMenuElem, BotMenuElemAttrText
from telegram_django_bot.routing import all_command_bme_handler, all_callback_bme_handler
from test_app.models import User
from django.conf import settings
import unittest


class BMEInitData:
    def setup_data(self, with_users=False, with_translation=False):
        self.bme_start = BotMenuElem.objects.create(
            command='start',
            callbacks_db='["start"]',
            message='Aloha, world!',
            buttons_db='[[{"text": "start", "callback_data": "start"}]]'
        )
        self.bme_2 = BotMenuElem.objects.create(
            command='second',
            callbacks_db='["second"]',
            message='Some message',
            buttons_db='['
                       '[{"text": "button 1", "callback_data": "button1"}],'
                       '[{"text": "button 2", "callback_data": "button2"}]'
                       ']'
        )
        if with_users:
            user_id = settings.TELEGRAM_TEST_USER_IDS[0]
            self.user = User.objects.create(id=user_id, username=user_id)

        if with_translation:
            BotMenuElemAttrText.objects.filter(
                bot_menu_elem=self.bme_start,
                language_code='ru',
                default_text='Aloha, world!'
            ).update(translated_text='Алоха, мир!')

            BotMenuElemAttrText.objects.filter(
                bot_menu_elem=self.bme_start,
                language_code='ru',
                default_text='start'
            ).update(translated_text='старт')


class TestBMEModels(TD_TestCase, BMEInitData):

    def setUp(self) -> None:
        self.setup_data()

    def test_translation_creating(self):
        self.assertEqual(10, BotMenuElemAttrText.objects.count())  # german + russian
        self.assertEqual(5, BotMenuElemAttrText.objects.filter(language_code='ru').count())
        self.assertEqual(4, BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_start).count())
        self.assertEqual(6, BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_2).count())

        self.bme_2.message = 'new message'
        self.bme_2.save()
        self.assertEqual(8, BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_2).count())

        self.bme_2.save()
        self.assertEqual(8, BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_2).count())

        BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_2).delete()
        self.bme_2.save()
        self.assertEqual(6, BotMenuElemAttrText.objects.filter(bot_menu_elem=self.bme_2).count())

        self.bme_2.delete()
        self.assertEqual(4, BotMenuElemAttrText.objects.count())

    def test_translation_getting(self):
        self.assertEqual('Aloha, world!', self.bme_start.get_message())
        self.assertEqual('Some message', self.bme_2.get_message())

        should_be = InlineKeyboardMarkup([[InlineKeyboardButton('start', callback_data='start')]])
        self.assertEqual(should_be.to_dict(), InlineKeyboardMarkup(self.bme_start.get_buttons()).to_dict())

        self.assertEqual('Aloha, world!', self.bme_start.get_message('ru'))
        self.assertEqual(should_be.to_dict(), InlineKeyboardMarkup(self.bme_start.get_buttons('ru')).to_dict())

        should_be = InlineKeyboardMarkup([
            [InlineKeyboardButton('button 1', callback_data='button1')],
            [InlineKeyboardButton('button 2', callback_data='button2')],
        ])
        self.assertEqual(should_be,  InlineKeyboardMarkup(self.bme_2.get_buttons()))


        BotMenuElemAttrText.objects.filter(
            bot_menu_elem=self.bme_start,
            language_code='ru',
            default_text='Aloha, world!'
        ).update(translated_text='Алоха, мир!')

        BotMenuElemAttrText.objects.filter(
            bot_menu_elem=self.bme_start,
            language_code='ru',
            default_text='start'
        ).update(translated_text='старт')

        self.assertEqual('Алоха, мир!', self.bme_start.get_message('ru'))
        should_be = InlineKeyboardMarkup([[InlineKeyboardButton('старт', callback_data='start')]])
        self.assertEqual(should_be.to_dict(), InlineKeyboardMarkup(self.bme_start.get_buttons('ru')).to_dict())

        self.assertEqual('Aloha, world!', self.bme_start.get_message('de'))
        should_be = InlineKeyboardMarkup([[InlineKeyboardButton('start', callback_data='start')]])
        self.assertEqual(should_be.to_dict(), InlineKeyboardMarkup(self.bme_start.get_buttons('de')).to_dict())


@unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
class TestBMEHandlers(TD_TestCase, BMEInitData):
    def setUp(self) -> None:
        self.setup_data(True, True)

    def test_command_handler(self):
        message = all_command_bme_handler(self.create_update({'text': '/start'}), self.test_callback_context)
        self.assertEqual('Aloha, world!', message.text)

        message = all_command_bme_handler(self.create_update({'text': '/second'}), self.test_callback_context)
        self.assertEqual('Some message', message.text)

        self.user.telegram_language_code = 'ru'
        self.user.save()
        message = all_command_bme_handler(self.create_update({'text': '/start'}), self.test_callback_context)
        self.assertEqual('Алоха, мир!', message.text)

        self.user.telegram_language_code = 'de'
        self.user.save()
        message = all_command_bme_handler(self.create_update({'text': '/start'}), self.test_callback_context)
        self.assertEqual('Aloha, world!', message.text)

    def test_command_handler_special_start(self):
        message = all_command_bme_handler(self.create_update({'text': '/start special'}), self.test_callback_context)
        self.assertEqual('Aloha, world!', message.text)

        self.bme_extra = BotMenuElem.objects.create(
            command='start special',
            callbacks_db='[]',
            message='extra start',
            buttons_db='[[{"text": "start", "callback_data": "start"}]]'
        )

        message = all_command_bme_handler(self.create_update({'text': '/start special'}), self.test_callback_context)
        self.assertEqual('extra start', message.text)

    def test_command_handler_no_command(self):
        message = all_command_bme_handler(self.create_update({'text': '/not_exist'}), self.test_callback_context)
        self.assertEqual(
            'Oops! It seems that an error has occurred, please write to support (contact in bio)!',
            message.text
        )

    def test_callback_handler(self):
        message = self.test_callback_context.bot.send_message(self.user.id, 'for test')
        message_params = {'message_id': message.message_id}
        callback_params = {'data': 'start'}

        message = all_callback_bme_handler(
            self.create_update(message_params, callback_params),
            self.test_callback_context
        )
        self.assertEqual('Aloha, world!', message.text)






from django.conf import settings

from django.utils import timezone
from django.db.models import Count
from django.db.models.functions import Coalesce
from django.db.models import OuterRef, Exists

from django.contrib.auth import get_user_model
from .models import Trigger, UserTrigger, ActionLog
from .tg_dj_bot import TG_DJ_Bot
from celery import current_app


@current_app.task
def create_triggers():
    def get_duration_dict(trigger, elem, add_prefix=True):
        prefix = 'actionlog__'
        if not add_prefix:
            prefix = ''
        return {
            f'{prefix}dttm__gte': dttm_now - trigger.get_timedelta(elem.pop('duration'))
        }

    dttm_now = timezone.now()
    User = get_user_model()

    for trigger in Trigger.objects.bot_filter_active().order_by('-priority').select_related('botmenuelem'):
        users = User.objects.filter(is_active=True)

        seeds = trigger.condition.pop('seeds', [])
        if seeds:
            users = users.filter(seed_code__in=seeds)

        # sequence = trigger.condition.pop('sequence', [])
        # if len(sequence):
        #     query = Q()
        #     for seq_elem in sequence:
        #         query = Q()
        #         for elem in seq_elem:
        #             kwargs = get_duration_dict(trigger, elem)
        #             for key, value in elem.items():
        #                 kwargs[f'actionlog__{key}'] = value
        #             query |= Q(**kwargs)
        #     users = users.filter(query)

        # exclude = trigger.condition.pop('exclude', [])
        # if len(exclude):
        #     query = Q()
        #     for elem in exclude:
        #         kwargs = get_duration_dict(trigger, elem)
        #         for key, value in elem.items():
        #             kwargs[f'actionlog__{key}'] = value
        #         query |= Q(**kwargs)
        #     users = users.exclude(query)

        check_amount = trigger.condition.pop('amount', [])
        if len(check_amount):
            for it, elem in enumerate(check_amount):
                filter_key = None
                filter_kwargs = get_duration_dict(trigger, elem, False)
                for key_d, value in elem.items():
                    if 'type' in key_d:
                        filter_kwargs[key_d] = value
                    else:
                        filter_key = key_d
                        filter_value = value
                if filter_key:
                    key = f'amount{it}'
                    subquery = ActionLog.objects.filter(
                        user_id=OuterRef('id'),
                        **filter_kwargs,
                    ).values('user_id').annotate(amount=Count('user_id')).values('amount')

                    users = users.annotate(**{
                        key: Coalesce(subquery, 0)
                    }).filter(**{f'{key}__{filter_key}': filter_value})

        user_triggers = UserTrigger.objects.filter(
            trigger=trigger,
            user=OuterRef('id'),
            dttm_added__gte=dttm_now - trigger.min_duration
        )

        users = users.annotate(exist_trigger=Exists(user_triggers)).filter(exist_trigger=False)

        # todo: intersection with other triggers -- minimum time between triggers to do?

        UserTrigger.objects.bulk_create([
            UserTrigger(trigger=trigger, user=user) for user in users
        ])

        trig_users = []
        for user in users:
            trig_users.append(user.id)
            if len(trig_users) == 500:
                send_triggers.delay(trig_users)
                trig_users = []

        if len(trig_users):
            send_triggers.delay(trig_users)



@current_app.task
def send_triggers(user_ids):
    dttm_now = timezone.now()
    UserTrigger.objects.filter(
        is_sent=False,
        user_id__in=user_ids,
        dttm_added__lt=dttm_now - timezone.timedelta(hours=16),  # triggers not sent in the first 16 hours are deleted
    ).update(dttm_deleted=dttm_now)

    user_triggers = UserTrigger.objects.filter(
        is_sent=False,
        dttm_deleted__isnull=True,
        user_id__in=user_ids,
    ).select_related('trigger', 'trigger__botmenuelem', 'user')

    bot = TG_DJ_Bot(settings.TELEGRAM_TOKEN)

    def _send_wrapper(self, *args, **kwargs):
        return bot.send_botmenuelem(*args, **kwargs)

    sent_user_triggers = []
    for user_trigger in user_triggers:
        is_sent, res_mess = bot.task_send_message_handler(
            _send_wrapper,
            user_trigger.user,
            None,
            user_trigger.user,  # for task_send_message_handler и для send_botmenuelem
            user_trigger.trigger.botmenuelem
        )
        if is_sent:
            sent_user_triggers.append(user_trigger)

    UserTrigger.objects.filter(id__in=[x.id for x in sent_user_triggers]).update(is_sent=True)
    ActionLog.objects.bulk_create([
        ActionLog(
            type=f'TRIGGER_SENT-{x.trigger_id}',
            user_id=x.user_id
        ) for x in sent_user_triggers
    ])





from telegram_django_bot.test import TD_TestCase
from telegram_django_bot.routing import telegram_reverse, RouterCallbackMessageCommandHandler
from django.conf import settings

from test_app.models import Entity, Order, Size, User, Category
from test_app.views import CategoryViewSet, EntityViewSet, OrderViewSet

import unittest
import telegram


class TestTelegramViewSet(TD_TestCase):
    def setUp(self) -> None:
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        self.user = User.objects.create(id=user_id, username=user_id)
        self.cvs = CategoryViewSet(
            telegram_reverse('CategoryViewSet'),
            user=self.user,
            bot=self.test_callback_context.bot
        )
        self.evs = EntityViewSet(
            telegram_reverse('EntityViewSet'),
            user=self.user,
            bot=self.test_callback_context.bot
        )
        self.ovs = OrderViewSet(
            telegram_reverse('OrderViewSet'),
            user=self.user,
            bot=self.test_callback_context.bot
        )
        # self.rc_mch = RouterCallbackMessageCommandHandler()
        # self.handle_update = lambda update: self.rc_mch.handle_update(
        #     update, 'some_str', 'some_str', self.test_callback_context
        # )

    def create_size(self):
        size = Size.objects.create(
            name='XL',
        )

        return size

    def create_category(self):
        category = Category.objects.create(
            info='Test',
            name='test'
        )

        return category

    def create_entity(self, category: Category, size: Size):
        entity = Entity.objects.create(
            name='test',
            category_id=category.pk,
            author_id=self.user.id,
            price=322
        )
        entity.sizes.add(size, through_defaults={'amount': 1})
        entity.save()

        return entity

    def create_order(self, entity: Entity):
        order = Order.objects.create(
            info='test',
            status=Order.STATUS_COMPLETED
        )
        order.entities.add(entity)
        order.save()

        return order

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_category_permission(self):
        category = self.create_category()
        category.name = 'hats'
        category.save()
        category2 = self.create_category()

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_delete_category = self.create_update(
            res_message.to_dict(),
            {'data': f'cat/de&{category.pk}'}
        )
        res_message = self.handle_update(action_delete_category)
        self.assertEqual(res_message.text, 'Sorry, you do not have permissions to this action.')

        action_delete_category2 = self.create_update(
            res_message.to_dict(),
            {'data': f'cat/de&{category2.pk}'}
        )
        res_message = self.handle_update(action_delete_category2)
        self.assertEqual(res_message.text, f'Are you sure you want to delete Category  #{category2.pk}?')

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)
        
        action_update_name = self.create_update(
            res_message.to_dict(),
            {'data': f'cat/up&{category.pk}&name'}
        )
        res_message = self.handle_update(action_update_name)
        self.assertEqual(res_message.text, 'Sorry, you do not have permissions to this action.')

        action_update_name2 = self.create_update(
            res_message.to_dict(),
            {'data': f'cat/up&{category2.pk}&name'}
        )
        res_message = self.handle_update(action_update_name2)
        self.assertEqual(res_message.text, 'Please, fill the field Category name\n\nmax length 128')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_show_list_order_with_foreign_filters(self):
        category = self.create_category()
        size = self.create_size()
        entity_1 = self.create_entity(category, size)
        entity_2 = self.create_entity(category, size)
        order_1 = self.create_order(entity_1)
        order_2 = self.create_order(entity_2)
        
        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_show_list = self.create_update(
            res_message.to_dict(),
            {'data': f'ord/sl&{entity_2.pk}'}
        )
        res_message = self.handle_update(action_show_list)

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], f'1. Order #{entity_2.pk}')
        self.assertEqual(buttons[0][0]['callback_data'], f'ord/se&{order_2.pk}&{entity_2.pk}')

    def test_create_order_v1(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        self.ovs.create('info', 'test')
        self.ovs.create('entities', f'{entity.pk}')
        self.ovs.create('entities', '!GNMS!')

        order_object = Order.objects.get(info='test', entities__pk=entity.pk)
        self.assertEqual(order_object.info, 'test')
        self.assertEqual(order_object.entities.all()[0].pk, entity.pk)

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_create_order_v2(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_create_order = self.create_update(
            res_message.to_dict(),
            {'data': 'ord/cr&&info&test'}
        )
        res_message = self.handle_update(action_create_order)

        action_add_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ord/cr&&entities&{entity.pk}'}
        )
        res_message = self.handle_update(action_add_entity)

        action_confirm_add_entity = self.create_update(
            res_message.to_dict(),
            {'data': 'ord/cr&&entities&!GNMS!'}
        )
        res_message = self.handle_update(action_confirm_add_entity)

        order_object = Order.objects.get(info='test', entities=entity.pk)

        mess = res_message.text.splitlines()
        self.assertEqual(mess[0], 'The Order is created! ')
        self.assertEqual(mess[2], f'Order #{order_object.pk} ')
        self.assertEqual(mess[3], f'Title: {order_object.info}')
        self.assertEqual(mess[4], f'Entities: {entity.name}')

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], '🔄 Title')
        self.assertEqual(buttons[0][0]['callback_data'], f'ord/up&&{order_object.pk}&info')

        self.assertEqual(buttons[0][1]['text'], '🔄 Entities')
        self.assertEqual(buttons[0][1]['callback_data'], f'ord/up&&{order_object.pk}&entities')

        self.assertEqual(buttons[1][0]['text'], f'❌ Delete #{order_object.pk}')
        self.assertEqual(buttons[1][0]['callback_data'], f'ord/de&&{order_object.pk}')

        self.assertEqual(buttons[2][0]['text'], '🔙 Return to list')
        self.assertEqual(buttons[2][0]['callback_data'], 'ord/sl&')

    def test_delete_entity_v1(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        _, (mess, buttons) = self.evs.delete(entity.pk)
        self.assertEqual(mess, f'Are you sure you want to delete Entity  #{entity.pk}?')

        button_1 = buttons[0][0].to_dict()
        self.assertEqual(button_1['text'], '🗑 Yes, delete')
        self.assertEqual(button_1['callback_data'], f'ent/de&{entity.pk}&{entity.pk}')

        button_2 = buttons[1][0].to_dict()
        self.assertEqual(button_2['text'], '🔙 Back')
        self.assertEqual(button_2['callback_data'], f'ent/se&{entity.pk}')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_delete_entity_v2(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_delete_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/de&{entity.pk}'}
        )
        res_message = self.handle_update(action_delete_entity)

        deletion_confirmation = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/de&{entity.pk}&{entity.pk}'}
        )
        res_message = self.handle_update(deletion_confirmation)

        message = res_message.text
        self.assertEqual(message, 'The Entity  #None is successfully deleted.')

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], '🔙 Return to list')
        self.assertEqual(buttons[0][0]['callback_data'], 'ent/sl')

        entity = Entity.objects.filter(category_id=category.pk).first()
        self.assertEqual(entity, None)

    def test_change_entity_v1(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)
        new_name = 'test_1'

        _, (mess, buttons) = self.evs.change(entity.pk, 'name', new_name)
        self.assertEqual(len(buttons), 5)

        mess = mess.splitlines()
        self.assertEqual(mess[0], 'The field has been updated!')
        self.assertEqual(mess[2], f'Entity #{entity.pk} ')
        self.assertEqual(mess[3], f'<b>Title</b>: {new_name}')
        self.assertEqual(mess[4], f'<b>Category</b>: {entity.category.name}')
        self.assertEqual(mess[5], f'<b>Sizes</b>: {entity.sizes.all()[0].name}')
        self.assertEqual(mess[6], f'<b>Is visable</b>: {entity.is_visable}')
        self.assertEqual(mess[7], f'<b>Price</b>: {entity.price:.2f}')
        self.assertEqual(mess[8], f'<b>Author id</b>: {self.user.id}')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_change_entity_v2(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)
        new_name = 'text_1'

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_change_name_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/up&{entity.pk}&name'}
        )
        res_message = self.handle_update(action_change_name_entity)

        message = res_message.text
        self.assertEqual(message, 'Please, write the value for field Title \n\nmax length 128')

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], '⬅️ Go back')
        self.assertEqual(buttons[0][0]['callback_data'], f'ent/se&{entity.pk}')

        write_name_for_entity = self.create_update({'text': new_name})
        res_message = self.handle_update(write_name_for_entity)

        mess = res_message.text.splitlines()
        self.assertEqual(mess[0], 'The field has been updated!')
        self.assertEqual(mess[2], f'Entity #{entity.pk} ')
        self.assertEqual(mess[3], f'Title: {new_name}')
        self.assertEqual(mess[4], f'Category: {entity.category.name}')
        self.assertEqual(mess[5], f'Sizes: {entity.sizes.all()[0].name}')
        self.assertEqual(mess[6], f'Is visable: {entity.is_visable}')
        self.assertEqual(mess[7], f'Price: {entity.price:.2f}')
        self.assertEqual(mess[8], f'Author id: {self.user.id}')

    def test_show_list_entity_v1(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        _, (mess, buttons) = self.evs.show_list()
        self.assertEqual(len(buttons), 1)

        button_1 = buttons[0][0].to_dict()
        self.assertEqual(button_1['text'], f'1. Entity #{entity.pk}')
        self.assertEqual(button_1['callback_data'], f'ent/se&{entity.pk}')

        mess = mess.splitlines()
        self.assertEqual(mess[0], f'1. Entity #{entity.pk}')
        self.assertEqual(mess[1], f'<b>Title</b>: {entity.name}')
        self.assertEqual(mess[2], f'<b>Category</b>: {entity.category.name}')
        self.assertEqual(mess[3], f'<b>Sizes</b>: {entity.sizes.all()[0].name}')
        self.assertEqual(mess[4], f'<b>Is visable</b>: {entity.is_visable}')
        self.assertEqual(mess[5], f'<b>Price</b>: {entity.price:.2f}')
        self.assertEqual(mess[6], f'<b>Author id</b>: {self.user.id}')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_show_list_entity_v2(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_show_list_entity = self.create_update(
            res_message.to_dict(),
            {'data': 'ent/sl'}
        )
        res_message = self.handle_update(action_show_list_entity)

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], f'1. Entity #{entity.pk}')
        self.assertEqual(buttons[0][0]['callback_data'], f'ent/se&{entity.pk}')

    def test_show_elem_entity_v1(self):
        category = self.create_category()
        size = self.create_size()
        entity = self.create_entity(category, size)

        _, (mess, buttons) = self.evs.show_elem(entity.pk)
        self.assertEqual(len(buttons), 5)

        mess = mess.splitlines()
        self.assertEqual(mess[0], f'Entity #{entity.pk} ')
        self.assertEqual(mess[1], f'<b>Title</b>: {entity.name}')
        self.assertEqual(mess[2], f'<b>Category</b>: {entity.category.name}')
        self.assertEqual(mess[3], f'<b>Sizes</b>: {entity.sizes.all()[0].name}')
        self.assertEqual(mess[4], f'<b>Is visable</b>: {entity.is_visable}')
        self.assertEqual(mess[5], f'<b>Price</b>: {entity.price:.2f}')
        self.assertEqual(mess[6], f'<b>Author id</b>: {self.user.id}')

        button_1 = buttons[0][0].to_dict()
        self.assertEqual(button_1['text'], '🔄 Title')
        self.assertEqual(button_1['callback_data'], f'ent/up&{entity.pk}&name')

        button_2 = buttons[0][1].to_dict()
        self.assertEqual(button_2['text'], '🔄 Category')
        self.assertEqual(button_2['callback_data'], f'ent/up&{entity.pk}&category')        

        button_3 = buttons[1][0].to_dict()
        self.assertEqual(button_3['text'], '🔄 Sizes')
        self.assertEqual(button_3['callback_data'], f'ent/up&{entity.pk}&sizes')

        button_4 = buttons[1][1].to_dict()
        self.assertEqual(button_4['text'], '🔄 Is visable')
        self.assertEqual(button_4['callback_data'], f'ent/up&{entity.pk}&is_visable')

        button_5 = buttons[2][0].to_dict()
        self.assertEqual(button_5['text'], '🔄 Price')
        self.assertEqual(button_5['callback_data'], f'ent/up&{entity.pk}&price')

        button_6 = buttons[2][1].to_dict()
        self.assertEqual(button_6['text'], '🔄 Author id')
        self.assertEqual(button_6['callback_data'], f'ent/up&{entity.pk}&author_id')

        button_7 = buttons[3][0].to_dict()
        self.assertEqual(button_7['text'], f'❌ Delete #{entity.pk}')
        self.assertEqual(button_7['callback_data'], f'ent/de&{entity.pk}')

        button_8 = buttons[4][0].to_dict()
        self.assertEqual(button_8['text'], '🔙 Return to list')
        self.assertEqual(button_8['callback_data'], 'ent/sl')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_show_elem_entity_v2(self):
        category = self.create_category()
        size = self.create_size()
        entity_object = self.create_entity(category, size)

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_show_elem_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/se&{entity_object.pk}'}
        )
        res_message = self.handle_update(action_show_elem_entity)

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], '🔄 Title')
        self.assertEqual(buttons[0][0]['callback_data'], f'ent/up&{entity_object.pk}&name')

        self.assertEqual(buttons[0][1]['text'], '🔄 Category')
        self.assertEqual(buttons[0][1]['callback_data'], f'ent/up&{entity_object.pk}&category')

        self.assertEqual(buttons[1][0]['text'], '🔄 Sizes')
        self.assertEqual(buttons[1][0]['callback_data'], f'ent/up&{entity_object.pk}&sizes')

        self.assertEqual(buttons[1][1]['text'], '🔄 Is visable')
        self.assertEqual(buttons[1][1]['callback_data'], f'ent/up&{entity_object.pk}&is_visable')

        self.assertEqual(buttons[2][0]['text'], '🔄 Price')
        self.assertEqual(buttons[2][0]['callback_data'], f'ent/up&{entity_object.pk}&price')

        self.assertEqual(buttons[2][1]['text'], '🔄 Author id')
        self.assertEqual(buttons[2][1]['callback_data'], f'ent/up&{entity_object.pk}&author_id')

        self.assertEqual(buttons[3][0]['text'], f'❌ Delete #{entity_object.pk}')
        self.assertEqual(buttons[3][0]['callback_data'], f'ent/de&{entity_object.pk}')

        self.assertEqual(buttons[4][0]['text'], '🔙 Return to list')
        self.assertEqual(buttons[4][0]['callback_data'], 'ent/sl')

    def test_create_entity_v1(self):
        category = self.create_category()
        size = self.create_size()
        self.evs.create('author_id', 1)
        self.evs.create('sizes', f'{size.pk}')
        self.evs.create('sizes', '!GNMS!')
        self.evs.create('name', 'test')
        self.evs.create('category', category.pk)
        self.evs.create('is_visable', '!NoneNULL!')
        self.evs.create('price', 322)

        entity = Entity.objects.get(
            category_id=category.pk,
            name='test'
        )
        self.assertEqual(entity.name, 'test')
        self.assertEqual(entity.price, 322)
        self.assertEqual(entity.category.pk, category.pk)
        self.assertEqual(entity.author_id, 1)

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_create_entity_v2(self):
        size_object = self.create_size()
        category_object = self.create_category()

        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_create_entity = self.create_update(
            res_message.to_dict(),
            {'data': 'ent/cr&price&322'}
        )
        res_message = self.handle_update(action_create_entity)

        write_name = self.create_update({'text': 'test'})
        res_message = self.handle_update(write_name)

        add_category_in_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/cr&category&{category_object.pk}'}
        )
        res_message = self.handle_update(add_category_in_entity)

        add_size_in_entity = self.create_update(
            res_message.to_dict(),
            {'data': f'ent/cr&sizes&{size_object.pk}'}
        )
        res_message = self.handle_update(add_size_in_entity)

        choose_size = self.create_update(
            res_message.to_dict(),
            {'data': 'ent/cr&sizes&!GNMS!'}
        )
        res_message = self.handle_update(choose_size)

        leave_blank = self.create_update(
            res_message.to_dict(),
            {'data': 'ent/cr&is_visable&!NoneNULL!'}
        )
        res_message = self.handle_update(leave_blank)

        write_author_id = self.create_update(
            res_message.to_dict(),
            {'data': 'ent/cr&author_id&12'}
        )
        res_message = self.handle_update(write_author_id)

        entity_object = Entity.objects.get(
            category_id=category_object.pk,
            name='test'
        )

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(len(buttons), 5)

        self.assertEqual(buttons[0][0]['text'], '🔄 Title')
        self.assertEqual(buttons[0][0]['callback_data'], f'ent/up&{entity_object.pk}&name')

        self.assertEqual(buttons[0][1]['text'], '🔄 Category')
        self.assertEqual(buttons[0][1]['callback_data'], f'ent/up&{entity_object.pk}&category')

        self.assertEqual(buttons[1][0]['text'], '🔄 Sizes')
        self.assertEqual(buttons[1][0]['callback_data'], f'ent/up&{entity_object.pk}&sizes')

        self.assertEqual(buttons[1][1]['text'], '🔄 Is visable')
        self.assertEqual(buttons[1][1]['callback_data'], f'ent/up&{entity_object.pk}&is_visable')

        self.assertEqual(buttons[2][0]['text'], '🔄 Price')
        self.assertEqual(buttons[2][0]['callback_data'], f'ent/up&{entity_object.pk}&price')

        self.assertEqual(buttons[2][1]['text'], '🔄 Author id')
        self.assertEqual(buttons[2][1]['callback_data'], f'ent/up&{entity_object.pk}&author_id')

        self.assertEqual(buttons[3][0]['text'], f'❌ Delete #{entity_object.pk}')
        self.assertEqual(buttons[3][0]['callback_data'], f'ent/de&{entity_object.pk}')

        self.assertEqual(buttons[4][0]['text'], '🔙 Return to list')
        self.assertEqual(buttons[4][0]['callback_data'], 'ent/sl')

    def test_create_category_v1(self):
        self.cvs.create('name', 'test')
        category = Category.objects.get(name='test')

        self.assertEqual(category.name, 'test')

    @unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
    def test_create_category_v2(self):
        category_name = 'шапки'

        start_send = self.create_update({'text': '/start'})
        res_message = self.handle_update(start_send)

        action_add_category = self.create_update(
            res_message.to_dict(),
            {'data': f'cat/cr&name&{category_name}'}
        )
        res_message = self.handle_update(action_add_category)

        category_object = Category.objects.get(name=category_name)
        self.assertEqual(category_name, category_object.name)

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(len(buttons), 3)

        button_1, button_2, button_3 = buttons
        self.assertEqual(button_1[0]['text'], '🔄 Category name')
        self.assertEqual(button_1[0]['callback_data'], f'cat/up&{category_object.pk}&name')

        self.assertEqual(button_2[0]['text'], f'❌ Delete #{category_object.pk}')
        self.assertEqual(button_2[0]['callback_data'], f'cat/de&{category_object.pk}')

        self.assertEqual(button_3[0]['text'], '🔙 Return to list')
        self.assertEqual(button_3[0]['callback_data'], 'cat/sl')

    def test_generate_links(self):
        self.assertEqual(
            'cat/cr&name&hat',
            self.cvs.gm_callback_data('create', 'name', 'hat')
        )

    def test_gm_delete_getting_confirmation(self):
        model = self.create_category()
        mess, buttons = self.cvs.gm_delete_getting_confirmation(model)
        button_1, button_2 = buttons

        button_1 = button_1[0].to_dict()
        self.assertEqual(button_1['text'], '🗑 Yes, delete')
        self.assertEqual(button_1['callback_data'], f'cat/de&{model.pk}&{model.pk}')

        button_2 = button_2[0].to_dict()
        self.assertEqual(button_2['text'], '🔙 Back')
        self.assertEqual(button_2['callback_data'], f'cat/se&{model.pk}')

        self.assertEqual(mess, 'Are you sure you want to delete Category  #1?')

    def test_gm_delete_successfully(self):
        model = self.create_category()
        mess, button = self.cvs.gm_delete_successfully(model)

        button = button[0][0].to_dict()
        self.assertEqual(button['text'], '🔙 Return to list')
        self.assertEqual(button['callback_data'], 'cat/sl')

        self.assertEqual(mess, f'The Category  #{model.pk} is successfully deleted.')

    def test_gm_no_elem(self):
        model = self.create_category()
        _, (mess, buttons) = self.cvs.gm_no_elem(model.pk)

        self.assertEqual(mess, f'The Category {model.pk} has not been found 😱 \nPlease try again from the beginning.')

    def test_gm_show_elem_create_buttons(self):
        model = self.create_category()
        button_1, button_2, button_3 = self.cvs.gm_show_elem_create_buttons(model)

        button_1 = button_1[0].to_dict()
        self.assertEqual(button_1['text'], '🔄 Category name')
        self.assertEqual(button_1['callback_data'], f'cat/up&{model.pk}&name')

        button_2 = button_2[0].to_dict()
        self.assertEqual(button_2['text'], f'❌ Delete #{model.pk}')
        self.assertEqual(button_2['callback_data'], f'cat/de&{model.pk}')

        button_3 = button_3[0].to_dict()
        self.assertEqual(button_3['text'], '🔙 Return to list')
        self.assertEqual(button_3['callback_data'], 'cat/sl')

    def test_gm_show_elem_or_list_fields(self):
        model = self.create_category()
        mess = self.cvs.gm_show_elem_or_list_fields(model)

        self.assertEqual(mess, f'<b>Category name</b>: {model.name}\n')

    def test_gm_value_str(self):
        model = self.create_category()
        value = self.cvs.gm_value_str(model, model.info, 'info')

        self.assertEqual(value, model.info)

    def test_gm_show_list_elem_info(self):
        model = self.create_category()
        mess = self.cvs.gm_show_list_elem_info(model, 1)

        self.assertEqual(mess, f'1. Category #{model.pk}\n<b>Category name</b>: {model.name}\n\n\n')

    def test_gm_show_list_button_names(self):
        model = self.create_category()
        mess = self.cvs.gm_show_list_button_names(1, model)

        self.assertEqual(mess, f'1. Category #{model.pk}')

    def test_gm_show_list_create_pagination(self):
        button_1, button_2 = self.cvs.gm_show_list_create_pagination(0, 10, 1, 2, 5)[0]

        button_1 = button_1.to_dict()
        self.assertEqual(button_1['text'], '◀️️️')
        self.assertEqual(button_1['callback_data'], 'cat/sl&-1')

        button_2 = button_2.to_dict()

        self.assertEqual(button_2['text'], '️▶️️')
        self.assertEqual(button_2['callback_data'], 'cat/sl&1')

    def test_gm_success_created(self):
        model = self.create_category()
        _, (mess, buttons) = self.cvs.gm_success_created(model.pk)
        button_1, button_2, button_3 = buttons

        self.assertEqual(mess, f'The Category is created! \n\nCategory #{model.pk} \n<b>Category name</b>: {model.name}\n')
        
        button_1 = button_1[0].to_dict()
        self.assertEqual(button_1['text'], '🔄 Category name')
        self.assertEqual(button_1['callback_data'], f'cat/up&{model.pk}&name')

        button_2 = button_2[0].to_dict()
        self.assertEqual(button_2['text'], f'❌ Delete #{model.pk}')
        self.assertEqual(button_2['callback_data'], f'cat/de&{model.pk}')

        button_3 = button_3[0].to_dict()
        self.assertEqual(button_3['text'], '🔙 Return to list')
        self.assertEqual(button_3['callback_data'], 'cat/sl')

    def test_gm_value_error(self):
        _, (mess, buttons) = self.cvs.gm_value_error('info', 'Error')
        button_1, button_2 = buttons

        button_1 = button_1[0].to_dict()
        self.assertEqual(button_1['text'], 'Leave blank')
        self.assertEqual(button_1['callback_data'], 'cat/cr&info&!NoneNULL!')

        button_2 = button_2[0].to_dict()
        self.assertEqual(button_2['text'], 'Cancel adding')
        self.assertEqual(button_2['callback_data'], 'cat/sl')

        self.assertEqual(mess, 'While adding Info the next errors were occurred: Error\n\nPlease, write the value for field Info \n\nextra info about category\n\n')

    def test_gm_self_variant(self):
        _, (mess, buttons) = self.cvs.gm_self_variant('info')
        button_1, button_2 = buttons

        button_1 = button_1[0].to_dict()
        self.assertEqual(button_1['text'], 'Leave blank')
        self.assertEqual(button_1['callback_data'], 'cat/cr&info&!NoneNULL!')

        button_2 = button_2[0].to_dict()
        self.assertEqual(button_2['text'], 'Cancel adding')
        self.assertEqual(button_2['callback_data'], 'cat/sl')

        self.assertEqual(mess, 'Please, write the value for field Info \n\nextra info about category\n\n')


@unittest.skipIf(int(telegram.__version__.split('.')[0]) >= 20, 'tests do not support async')
class TestUserViewSet(TD_TestCase):
    def setUp(self) -> None:
        user_id = settings.TELEGRAM_TEST_USER_IDS[0]
        self.user = User.objects.create(id=user_id, username=user_id)
        
        self.rc_mch = RouterCallbackMessageCommandHandler()
        self.handle_update = lambda update: self.rc_mch.handle_update(
            update, 'some_str', 'some_str', self.test_callback_context
        )

    def test_show_elem(self):
        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_show_settings = self.create_update(
            res_message.to_dict(),
            {'data': 'us/se'}
        )
        res_message = self.handle_update(action_show_settings)

        buttons = res_message.reply_markup.to_dict()['inline_keyboard']
        self.assertEqual(buttons[0][0]['text'], '🔄 Timezone')
        self.assertEqual(buttons[0][0]['callback_data'], f'us/up&{self.user.id}&timezone')
        self.assertEqual(buttons[0][1]['text'], '🔄 Language')
        self.assertEqual(buttons[0][1]['callback_data'], f'us/up&{self.user.id}&telegram_language_code')
        self.assertEqual(res_message.text, 'Timezone: +3 UTC\nLanguage: English')

    def test_change_language(self):
        start = self.create_update({'text': '/start'})
        res_message = self.handle_update(start)

        action_change_timezone = self.create_update(
            res_message.to_dict(),
            {'data': f'us/up&{self.user.id}&telegram_language_code'}
        )
        res_message = self.handle_update(action_change_timezone)
        
        message = res_message.text
        buttons = res_message.reply_markup.to_dict()['inline_keyboard']

        self.assertEqual(message, 'Please, fill the field Language')

        self.assertEqual(buttons[0][0]['text'], '✅ English')
        self.assertEqual(buttons[0][0]['callback_data'], f'us/up&{self.user.id}&telegram_language_code&en')

        self.assertEqual(buttons[1][0]['text'], 'German')
        self.assertEqual(buttons[1][0]['callback_data'], f'us/up&{self.user.id}&telegram_language_code&de')

        self.assertEqual(buttons[2][0]['text'], 'Russian')
        self.assertEqual(buttons[2][0]['callback_data'], f'us/up&{self.user.id}&telegram_language_code&ru')

        self.assertEqual(buttons[3][0]['text'], '⬅️ Go back')
        self.assertEqual(buttons[3][0]['callback_data'], f'us/se&{self.user.id}')
